#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>

#include "config.h"
#include "comm.h"


static unsigned char PacketDataBuffer[1024];	// receive buffer

int random(int range) {
  return (int) (range*1.0*rand()/(RAND_MAX+1.0));
}

TPacket Packet(void)				// return empty packet (but with *data set correctly to internal buffer)
{
  TPacket pac;
  pac.size = 0;
  pac.packetType = packetEmpty;
  pac.data = PacketDataBuffer+1;
  return pac;
}

TPacket RecvPacket(int fd, TPacketType pt)		// receive a packet of selected type (ignore any other packets) 
							// [blocking if fd>0] [nonblocking if fd<0, also ignores pt]
{
  TPacket packet;

  fd_set fds;
  int res;
  int cycle = (fd>0);
  fd=abs(fd);

  do {
    int have_packet=0;
    
    if (cycle) {
      FD_ZERO(&fds);
      FD_SET(fd,&fds);
    
      select(fd+1,&fds,NULL,NULL,NULL);		// wait for a packet
     
      if (FD_ISSET(fd,&fds)) have_packet=1;
    } else have_packet=1;
    
//      res = read(fd,PacketDataBuffer,sizeof(PacketDataBuffer));
    if (have_packet) {
      res = recv(fd,PacketDataBuffer,sizeof(PacketDataBuffer),0);
//      printf("RECV: res=%d, errno=%d, from=0x%08X\n",res,errno,*(PacketDataBuffer+1));
      if (res>0) {
        int act_type = (int)PacketDataBuffer[0];

//	printf("IN_packet: type=%d, len=%d, waiting for %d\n",act_type,res,pt);
//	for (int i=0; i<res;i++) printf("%s%02X ",(i%16==0 ? "\n  ":""),PacketDataBuffer[i]);
//	printf("\n---------- END OF PACKET\n");
      
        // only end this loop after correct packet arrives...
        if ((act_type == pt) || (!cycle)) {
          break;
        }
      } else {
        // this is an error -- res==0 means EOF (master died?), res==-1 means error
        printf("ERR: res=%d\n",res);
	exit(2+res);
      }
    }
    
  } while (cycle);
  
  packet.size = res;
  packet.packetType = pt;
  packet.data = PacketDataBuffer+1;
  
  return packet;
}

int SendPacket(int fd, TPacket packet)			// send a packet
{
  int res;
  
  PacketDataBuffer[0]=packet.packetType;
  switch (packet.packetType) {
    case packetEmpty:     packet.size=0; break;
    case packetBorn:      packet.size=21; break;
    case packetDoStep:    packet.size=GetByte(packet,10)*8 + 11 + 2 + MAP_SIZE*MAP_SIZE*4; break;
    case packetDoneStep:  packet.size=13; break;
    case packetQuit:      packet.size=0; break;
    case packetName:      packet.size=8+GetInt(packet,4); break;
    default:              packet.size=0; break;
  }

//  printf("OUT_packet: type=%d, len=%d\n",packet.packetType,packet.size);
//  for (int i=0; i<packet.size;i++) printf("%s%02X ",(i%16==0 ? "\n  ":""),PacketDataBuffer[i]);
//  printf("\n---------- END OF PACKET\n");


  res = write(fd,PacketDataBuffer,packet.size+1);
//  printf("SendPacket: res=%d, errno=%d\n",res,errno);
  return (res != -1);
}

int GetInt(TPacket pac, int offset)		// get int (4bytes) from specified offset
{
  int i;
  memcpy(&i,pac.data+offset,4);
  return i;
}

int GetByte(TPacket pac, int offset)		// get byte (1byte) from specified offset
{
  return pac.data[offset];
}

void SetInt(TPacket pac, int offset, int val)			// set int (4bytes) from specified offset
{
  memcpy(pac.data+offset,&val,4);
}

void SetByte(TPacket pac, int offset, unsigned char val)	// set byte (1byte) from specified offset
{
  pac.data[offset]=val;
}

