#ifndef _CONFIG_H
#define _CONFIG_H

#define DEBUG(x...) printf(x);

/*
 entities available:
    Base	(type 1)	// cannot move, can buy new fighters/workers, sugar storage
    Fighter	(type 2)	// kills another worker, with 50% probability another fighter
    Worker	(type 3)	// can collect sugar and transport it to the Base

 step:

   IN:  x,y coordinates within a map
        visible piece of map (square around x,y), possibly with another entities
        incomming mesasges (up to 5)

   OUT: movement (direction up/right/down/left)
        message to send
*/

// ------- public structures -----------------------------

// maximum number of players
#define MAX_PLAYERS 	6
#define MAX_ENT 	100

// map size is always odd number
#define MAP_SIZE 	7	

// full map size
#define GLOBAL_MAP_SIZE	105

// maximum number of input messages
#define MAX_MESSAGES 	5

// how much sugar can be taken from one field
#define SUGAR_ITEM	1

// probability that worker will kill fighter
#define PROB_KILL	5

enum TEntityType {etEmpty=0,  etBase=1, etFighter=2, etWorker=3,  etWall=4};

// what to do:
//  - nothing
//  - move in one of 4 directions
//  - send message to someone
//  - drop piece of sugar (only valid for workers with sugar)
enum TMoveAction {dirStay=0, dirUp=1, dirRight=2, dirDown=3, dirLeft=4, dirSend=5, dirDrop=6};

enum TBornEntity {bornNone=0, bornFighter=1, bornWorker=2};

#define MASK_ALL	0xFFFFFFFF

#define MASK_TYPE	0xF0000000
#define MASK_PLAYER     0x0F000000
#define MASK_SUGAR 	0x00FF0000
#define MASK_SEQ	0x0000FFFF

#define SHIFT_TYPE	28
#define SHIFT_PLAYER	24
#define SHIFT_SUGAR     16
#define SHIFT_SEQ	0
// USAGE:    (id & MASK_xyz) >> SHIFT_xyz


// id deconding macros

// ID_TYPE: return entity type (0..4 - TEntityEnum)
#define ID_TYPE(id)     ((id & MASK_TYPE) >> SHIFT_TYPE)

// ID_PLAYER: return player number (1..MAX, or 0 if no player at this field)
#define ID_PLAYER(id)   ((id & MASK_PLAYER) >> SHIFT_PLAYER)

// ID_SUGAR: return amount of sugar (0..255)
#define ID_SUGAR(id)    ((id & MASK_SUGAR) >> SHIFT_SUGAR)

// ID_SEQ: return entity sequence number (unique within single player)
#define ID_SEQ(id)      ((id & MASK_SEQ) >> SHIFT_SEQ)


// set parts of id
#define SET_TYPE(id,q)    id = (id & (MASK_ALL ^ MASK_TYPE))   | (q << SHIFT_TYPE);
#define SET_PLAYER(id,q)  id = (id & (MASK_ALL ^ MASK_PLAYER)) | (q << SHIFT_PLAYER);
#define SET_SUGAR(id,q)   id = (id & (MASK_ALL ^ MASK_SUGAR))  | (q << SHIFT_SUGAR);
#define SET_SEQ(id,q)     id = (id & (MASK_ALL ^ MASK_SEQ))    | (q << SHIFT_SEQ);


#define FONT_SCALE 	0.6
#define NR_VAULTS	2
#define RET_HALT	47

#define SCREEN_W 1024
#define SCREEN_H 768

//#define SCREEN_W 800
//#define SCREEN_H 600

//#define SCREEN_W 640
//#define SCREEN_H 480

//#define SCREEN_W 1280
//#define SCREEN_H 1024



#define MAX_TIMEOUT	20

typedef int TID;	// unique identifier:  (type << 24) | (sugar << 23) | (player << 16) | (player's entity sequence number)
typedef int TMsgData;	// data message

typedef struct {
  TID       id;		// ID of sender for in_messages, ID of recipient for out_messages
  TMsgData  msg;	// actual message
} TMessage;

typedef struct {
  TID field[MAP_SIZE][MAP_SIZE];	// visible map is square, you are always in the center
} TMap;

typedef struct {
  TID    id;		// unique identifier
  int    x,y;		// coordinates
  int 	 sugar;		// amount of sugar carried by this entity

  // in-message queue
  int      in_msgs;	// number of messages
  TMessage in_msg[MAX_MESSAGES];	// messages

  //  internal data
  int pid;		// pid of corresponding process
  int fd;		// unix socket to this process
  int pending;		// number of pending DoStep messages
  int timeout;		// how many times this entity wasn't called because pending>0
  int zombie;		// was I killed?
  int sprite;		// sprite index into obj_sprites
} TEntity;

typedef struct {
  int workers, fighters;
} TEntityCount;

typedef struct {
  int w,h;		// map size
  int maxSugar;		// max. amount of sugar per field
  int initSugar;	// amount of sugar in base at the beginning
  int prizeWorker, prizeFighter;	// prize of worker and fighter
  int maxSteps;		// max steps per game
  
  struct {
    int x,y;
  } bases[MAX_PLAYERS];	// base positions
  
  TID field[GLOBAL_MAP_SIZE][GLOBAL_MAP_SIZE];
} TGlobalMap;

int random(int range);

#endif  /* _CONFIG_H */
