#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <time.h>
#include <signal.h>

#include "config.h"
#include "comm.h"
#include "conffiles.h"


#include "gui.h"
#include "score.h"

#define COLOR(r,rr,g,gg,b,bb)   { color.SetRed(r,rr); color.SetGreen(g,gg); color.SetBlue(b,bb); }
#define DELAY(a,b) 		{ struct timeval tv; tv.tv_sec=a; tv.tv_usec=b; select(0,NULL,NULL,NULL,&tv); }


#define COL_NORM	"\033[00m"
#define COL_BRIGHT	"\033[01m"

#define COL_RED		"\033[31m"
#define COL_GREEN	"\033[32m"
#define COL_BLUE	"\033[34m"
#define COL_GRAY 	"\033[37m"

#define CUR_MOVE_HOME   "\033[H"


struct {
  char *exename;			// name of player's executable
  char *name;				// player name (set buring Born)

  TEntity ent[MAX_ENT];			// list of all entities
  int base;				// base entity -- index to ent[]
  int entCount;				// number of entities
  int seq;				// sequence number
  
//  int sugar;				// amount of sugar
  int sugar_sum;			// total sugar I've brought
  int kills;				// how many opponents I've killed
  int score_game;			// score per game
  
} Player[MAX_PLAYERS];			// info about players


// global scoring, ...
struct {
  int score;
} GlobalPlayer[MAX_PLAYERS];


int nPlayers;				// actual number of players in a game
TGlobalMap map;

int sugar_to_sprite[(GLOBAL_MAP_SIZE+10)*256];		// s_t_s[x<<8 | y] -- kde v obj_sprite[] je policko s cukrom [x][y]



SDL_Surface *screen;
KrEngine *engine;
KrTextBox **text;
KrResourceVault *vaults[NR_VAULTS];

KrSprite **obj_sprites;

int in_intro = 0;

FILE *logfile;


//---------------------------------------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------------------------------------

TPacket mkBornPacket(TEntity *ent, TGlobalMap *map) {
  TPacket pac = Packet();
  pac.packetType = packetBorn;

  int offset=0;
  SetInt (pac,offset,ent->id);     offset+=4;
  SetByte(pac,offset,map->w);      offset++;
  SetByte(pac,offset,map->h);      offset++;  
  SetByte(pac,offset,ent->x);      offset++;
  SetByte(pac,offset,ent->y);      offset++;
  SetByte(pac,offset,nPlayers);    offset++;
  SetInt (pac,offset,ent->sugar);          offset+=4;
  SetInt (pac,offset,map->prizeWorker);    offset+=4;
  SetInt (pac,offset,map->prizeFighter);   offset+=4;
  
  return pac;
}

TPacket mkDoStepPacket(TEntity *ent, TGlobalMap *map) {
  TPacket pac = Packet();
  pac.packetType = packetDoStep;
  int pl = ((ent->id & MASK_PLAYER) >> SHIFT_PLAYER)-1;
  int et = (ent->id & MASK_TYPE) >> SHIFT_TYPE;

  SetInt (pac,0,ent->id);
  SetByte(pac,4,ent->x);
  SetByte(pac,5,ent->y);
  SetInt (pac,6,ent->sugar);
  SetByte(pac,10,ent->in_msgs);
  int offset=11;
  
  // entity counts
  if (et==etBase) {
    int ec_worker=0, ec_fighter=0;
    for (int i=0; i<Player[pl].entCount; i++) {
      if ( ((Player[pl].ent[i].id & MASK_TYPE) >> SHIFT_TYPE) == etWorker) ec_worker++;
      if ( ((Player[pl].ent[i].id & MASK_TYPE) >> SHIFT_TYPE) == etFighter) ec_fighter++;
    }
    SetByte(pac,offset++,ec_worker);
    SetByte(pac,offset++,ec_fighter);
  } else {
    SetByte(pac,offset++,0);
    SetByte(pac,offset++,0);
  }
  
  // map
  int size2=(MAP_SIZE-1)/2;
  for (int j=-size2; j<=size2; j++)
   for (int i=-size2; i<=size2; i++) {
     int x=ent->x+i;
     int y=ent->y+j;
     if ((x>=0) && (x<map->w) && (y>=0) && (y<map->h)) SetInt(pac,offset,map->field[x][y]);
       else SetInt(pac,offset,etWall<<SHIFT_TYPE);
     offset+=4;
   }
  
  // messages
  for (int i=0; i<ent->in_msgs; i++) {
//    DEBUG("> SENDING MESSAGE: from 0x%08X, to 0x%08X, msg 0x%08X\n",ent->id, ent->in_msg[i].id, ent->in_msg[i].msg);
    SetInt(pac,offset,ent->in_msg[i].id);
    SetInt(pac,offset+4,ent->in_msg[i].msg);
    offset+=8;
  }
  
  return pac;
}

bool born_findPlace(int pl, int *x, int *y) {
  if (((*x)>=0) && ((*y)>=0)) return true;		// user-specified position
 
  if (Player[pl].base<0) return false; 
  int bx=Player[pl].ent[Player[pl].base].x;	// base location
  int by=Player[pl].ent[Player[pl].base].y;

  // find nearest empty place to the base
  int rad=1;
  while (rad<=(MAP_SIZE/2)) {
    for (int q=-rad; q<=rad; q++) {
      if (map.field[bx+q][by-rad] == 0) { *x=bx+q; *y=by-rad; return true; }
      if (map.field[bx+q][by+rad] == 0) { *x=bx+q; *y=by+rad; return true; }
      if (map.field[bx-rad][by+q] == 0) { *x=bx-rad; *y=by+q; return true; }
      if (map.field[bx+rad][by+q] == 0) { *x=bx+rad; *y=by+q; return true; }
    }
    rad++;
  }
  return false;
}

TEntity *born_entity(int pl, int enType, int x, int y, KrSprite **obj_sprites) 
{
  if (Player[pl].entCount<MAX_ENT) {
//    DEBUG("Born new entity: pl=%d, et=%d\n",pl,enType);

    TEntity *ent = &(Player[pl].ent[Player[pl].entCount]);

    ent->id = (enType << SHIFT_TYPE) | ((pl+1) << SHIFT_PLAYER) | (0 << SHIFT_SUGAR) | (Player[pl].seq++ << SHIFT_SEQ);
    if (!born_findPlace(pl, &x, &y)) return NULL;	// no place to create new entity :(

    ent->x = x;
    ent->y = y;
    
    ent->in_msgs = 0;
    ent->sugar = (enType==etBase ? map.initSugar : 0);
    ent->pending = 0;
    ent->zombie = 0;
    ent->sprite = 0;

    // create new process for player,
    int sv[2];
    int res = socketpair(PF_UNIX, SOCK_DGRAM, 0, sv);
    if (!res) {
      fcntl(sv[0],F_SETFD,FD_CLOEXEC);	// set sv[0] to be closed upon executing players
      ent->fd=sv[0];

      // start player process and connect it
      if ((ent->pid = fork())==0) {		// child
        // close all unused file descriptors...
        close(0);
//	close(1);
	close(2);
	
	// link created socket to fd 7
	dup2(sv[1],7);

	char s[128]; s[127]=0;
	char *TypeNames[] = {"", "Base", "Fighter", "Worker", "X"};
	snprintf(s,127,"%s (%s)",Player[pl].exename,TypeNames[enType]);
	
	int res = execlp(Player[pl].exename,(s[0]=='.' ? s+2 : s),NULL);
	
	DEBUG("exec 0x%08X failed: res=%d, errno=%d\n",ent->id,res,errno);
	exit(1);	// this is child -- STOP!!!
      } else {				// parent
//	DEBUG("child 0x%08X started: pid=%d, errno=%d\n",ent->id,ent->pid,errno);

	if (enType==etBase) Player[pl].base=Player[pl].entCount;

	
	switch (enType) {
	  case etBase:    ent->sprite = create_sprite(x,y,2,6); break;	// x, y, frame number, z-depth
	  case etFighter: ent->sprite = create_sprite(x,y,3,7); break;	// x, y, frame number, z-depth
	  case etWorker:  ent->sprite = create_sprite(x,y,4,7); break;	// x, y, frame number, z-depth
	}


	KrColorTransform color;
	COLOR(player_color[pl][0],0, player_color[pl][1],0, player_color[pl][2],0);
	obj_sprites[ent->sprite]->SetColor(color);


	Player[pl].entCount++;
	map.field[ent->x][ent->y] = ent->id;		// insert entity into map

        // send born info to player
	SendPacket(ent->fd, mkBornPacket(ent, &map));

        // wait for answer -- player name
	if (enType == etBase) {
	  DEBUG("Waiting for name packet from %d...\n",pl+1);
   	  TPacket pac = RecvPacket(ent->fd, packetName);
 	  int name_len = GetInt(pac,4);
	  Player[pl].name = (char *)malloc(name_len+1);  bzero(Player[pl].name, name_len+1);
	  for (int i=0; i<name_len; i++) Player[pl].name[i] = GetInt(pac,8+i);
	  DEBUG("OK, name is '%s'\n",Player[pl].name);
	}

	
        close(sv[1]);
      }
    } else {
      DEBUG("Failed to create socketpair: errno=%d\n",errno);
      return NULL;
    }

    return ent;
  } else 			// too many entities
    return NULL;
}

// ---------------------------------------------------------------------------
TEntity *findEntity(TID id)
{
  id = id & (MASK_ALL ^ MASK_SUGAR);		// there might be a SUGAR flag set --> clear it first
  int pl = ((id & MASK_PLAYER) >> SHIFT_PLAYER) -1;

  for (int i=0; i<Player[pl].entCount; i++)
    if (Player[pl].ent[i].id == id)  {
      return &(Player[pl].ent[i]);
    }

  return NULL;
}

// ---------------------------------------------------------------------------
// process DoneStep packet
//	 process_DoneStep(ent, pl, i, pac);
void process_DoneStep(TEntity *ent, int pl, TPacket pac, KrSprite **obj_sprites)
{
  // parse packet
  TID id = GetInt(pac,0);
  if (id!=ent->id) {
    DEBUG("Spoofed packed received! id=0x%08X, should be 0x%08X!\n",id,ent->id);
    return;
  }
  if (ent->zombie) return;		// zombies can't move...
  
  int dir = GetByte(pac,4);
  TMessage msg;
  msg.id = GetInt(pac,5);
  msg.msg = GetInt(pac,9);
  
  // dir tabs: Stay, 4xDir, Send, Drop
  int DirTabX[] = {0, 0, 1, 0, -1, 0, 0};
  int DirTabY[] = {0, 1, 0, -1, 0, 0, 0};

  int mt = (ent->id & MASK_TYPE) >> SHIFT_TYPE;		// my type 
  switch (mt) {
    case etBase:
	    // dir: if nonzero -> born new entity (TBornEntity)
    	    switch (dir) {
	      case bornFighter: if (ent->sugar >= map.prizeFighter)
	                          if (born_entity(pl,etFighter,-1,-1,obj_sprites) != NULL) ent->sugar-=map.prizeFighter;
	    			break;
				
	      case bornWorker:  if (ent->sugar >= map.prizeWorker)
	                          if (born_entity(pl,etWorker,-1,-1,obj_sprites) != NULL) ent->sugar-=map.prizeWorker;
	    			break;
	    }
	    break;
	    
    case etFighter:
    case etWorker:
	    //=============================================================================================================
	    if ((dir>=0) && (dir<=5)) {
	      int nx=ent->x+DirTabX[dir];
	      int ny=ent->y+DirTabY[dir];
	      int sug_old = ID_SUGAR(map.field[ent->x][ent->y]);
	      
	      int who_there = (map.field[nx][ny] & MASK_TYPE) >> SHIFT_TYPE;
	      int who_pl = ((map.field[nx][ny] & MASK_PLAYER) >> SHIFT_PLAYER)-1;
	      int sug_new = ID_SUGAR(map.field[nx][ny]);
	      
	      int kill[] = {0,0};	// if this move kills [0]=me, [1]=the one on field I want to go to
	      int canmove = 0;		// if I can move to the new field (not if W<->W)
	      
	      switch (who_there) {
	        case etEmpty:		// can go there directly
			canmove = 1;
			break;
			
		case etWorker :
		        switch (mt) {
			  case etWorker:	kill[0] = 0;
						kill[1] = 0;
						canmove = (!kill[0]) && (kill[1]);
						break;
						
			  case etFighter:	kill[0] = (who_pl!=pl) && (random(PROB_KILL)==0);
						kill[1] = (who_pl!=pl);
		    				canmove = (!kill[0]) && (kill[1]);
						break;
			}
			break;

		case etFighter:	
		        switch (mt) {
			  case etWorker:	kill[0] = (who_pl!=pl);
						kill[1] = (who_pl!=pl) && (random(PROB_KILL)==0);
						canmove = (!kill[0]) && (kill[1]);
						break;
						
			  case etFighter:	kill[0] = (who_pl!=pl);
						kill[1] = (who_pl!=pl);
		    				canmove = (!kill[0]) && (kill[1]);
						break;
			}
			break;
			
		case etWall:
		default:
			canmove = 0;			// I can't fly above the walls...
			break;
	      }

	      
	      // don't go from sugar to another sugar
	      if ( (sug_old) && (sug_new) ) {
	        canmove = 0;
		kill[0] = kill[1] = 0;
	      }
	      
	      // was somebody killed?
	      if (kill[0]) {				// if I was killed, clear me from the map and make a zombie of me
	        ent->zombie=1;
	      }
  
	      if (kill[1]) {				// the same with my opponent
	        TEntity *opp = findEntity(map.field[nx][ny]);
		if (opp!=NULL) opp->zombie=1;
		Player[pl].kills += ( who_there == etWorker ? map.prizeWorker : map.prizeFighter );
	      }
	      
	      // can I move to new field?
	      if (canmove) {	
	        map.field[nx][ny] = (map.field[nx][ny] & MASK_SUGAR) | ent->id;
		// don't clear sugar amount from the field if we don't move...
		if ((nx!=ent->x) || (ny!=ent->y)) map.field[ent->x][ent->y] &= MASK_SUGAR;
		ent->x = nx;
		ent->y = ny;
	      }
	      
	      // is there a sugar where I am standing (and mt==etWorker) (and don't have sugar already)?
	      if ((mt==etWorker) && (map.field[ent->x][ent->y] & MASK_SUGAR) && (ent->sugar==0)) {
	        ent->sugar++;
		
		// take away 1 sugar
		int sug=(map.field[ent->x][ent->y] & MASK_SUGAR) >> SHIFT_SUGAR;
		sug--;
		
		// store the result
		map.field[ent->x][ent->y] = (map.field[ent->x][ent->y] & (MASK_ALL ^ MASK_SUGAR)) | (sug << SHIFT_SUGAR);
		
		if (sug==0) {
		  delete_sprite(sugar_to_sprite[(ent->y << 8) | ent->x]);
		  sugar_to_sprite[(ent->y << 8) | ent->x] = 0;
		}
	      }
	      
	      // transfer sugar to base: if worker is near-by (its own;)) base
	      if (Player[pl].base>=0) { 
  	        int base_dx = abs(ent->x - Player[pl].ent[Player[pl].base].x);
	        int base_dy = abs(ent->y - Player[pl].ent[Player[pl].base].y);
	        if ((mt==etWorker) && (base_dx<=1) && (base_dy<=1) && (ent->sugar)) {
	          // transfer sugar to base...
		  Player[pl].ent[Player[pl].base].sugar += ent->sugar;
		  Player[pl].sugar_sum++;
		  ent->sugar=0;
	        }
	      }
	    }	// end: dir>=0 && dir<=5   (i.e. Stay, 4xDir, Send

	    //=============================================================================================================
	    if ((dir==dirDrop) && (ent->sugar) && (mt==etWorker)) {		// drop piece of sugar here
	      ent->sugar--;
	      // put sugar to map
	      int sug=(map.field[ent->x][ent->y] & MASK_SUGAR) >> SHIFT_SUGAR;
	      sug++;
	      if (sug>255) sug=255;
		
	      // store the result
	      map.field[ent->x][ent->y] = (map.field[ent->x][ent->y] & (MASK_ALL ^ MASK_SUGAR)) | (sug << SHIFT_SUGAR);
	    }   // end: dir==dirDrop
	    
	    //=============================================================================================================
	    break;	// mt == etWorker, etFighter
  }
  
  // process outgoing messages (not for Base -- see 'question 1')
  if ((msg.id!=0) && (mt!=etBase)){
    // check if we can send message to this id (square around NEW entity position)
    int size2=(MAP_SIZE-1)/2;
    int found=0;
    for (int j=-size2; j<=size2; j++) {
     for (int i=-size2; i<=size2; i++) {
       int x=ent->x+i;
       int y=ent->y+j;
       
       // if this is the recipient, insert msg into his message queue
       if ((x>=0) && (x<map.w) && (y>=0) && (y<map.h) && ((int)(map.field[x][y] & (MASK_ALL ^ MASK_SUGAR)) == msg.id)) {
         TEntity *rcpt = findEntity(msg.id);
         if (rcpt!=NULL) {
	   found=1;
	   
//	   DEBUG(">>+++++++++++++++++++++++++++++\n");
//	   DEBUG("Got message from 0x%08X to 0x%08X: msg=%d\n",ent->id, msg.id, msg.msg);
	   if (rcpt->in_msgs < MAX_MESSAGES) {
	     rcpt->in_msg[rcpt->in_msgs].id  = ent->id;		// from:
    	     rcpt->in_msg[rcpt->in_msgs].msg = msg.msg;		// subj:
	     rcpt->in_msgs++;
	   }
	 }
	 break;
       }
     }
     if (found) break;
    } 
  }
}

void LoadMap(char *fname, TGlobalMap *map, KrSprite ** /* objs */) 
{
  char s[1024];
  sprintf(s,"%s.map",fname);

  FILE *f = fopen(s,"r");
  if (f==NULL) return;

  int row=0;
  
  bzero(map,sizeof(map));
  
  while (!feof(f)) {
    bzero(s,sizeof(s));
    char *q = fgets(s,sizeof(s),f);
    if (q==NULL) break;
    
    if ((strlen(s)==1) && (s[0]=='\n')) continue;
    if ((q[0]==';') || (q[0]=='#')) continue;

    switch (row) {
      case 0: sscanf(s,"%d %d ",&(map->w), &(map->h));
              row++; break;
	      
      case 1: sscanf(s,"%d ",&(map->maxSugar));
              row++; break; 

      case 2: sscanf(s,"%d %d %d ", &(map->initSugar), &(map->prizeWorker), &(map->prizeFighter));
              row++; break; 

      case 3: sscanf(s,"%d ",&(map->maxSteps));
              row++; break; 

      case 4: for (int y=map->h-1; y>=0; y--) {
                if (y!=map->h-1) q=fgets(s,sizeof(s),f);
	        for (int x=0; x<map->w; x++) {
		  switch (q[x]) {
		    case 'x':
		    case 'X': map->field[x][y]=(etWall << SHIFT_TYPE); break;
		    case '.': map->field[x][y]=(etEmpty << SHIFT_TYPE); break;
		    
		    case '0': case '1': case '2': case '3': case '4':
		    case '5': case '6': case '7': case '8': case '9':
		              map->field[x][y] = (etEmpty << SHIFT_TYPE) | (int)((q[x]-'0'+1)*1.0/10*(map->maxSugar)) << SHIFT_SUGAR; 
			      break;
		  
		    case 'a': case 'b': case 'c': case 'd': case 'e':
		    case 'f': case 'g': case 'h': case 'i': case 'j':
		              map->bases[q[x]-'a'].x=x;
		              map->bases[q[x]-'a'].y=y;
			      break;
		  }
		} 	
	      }
	      
	      // set correct scaling to fit the screen
	      double sx=(1.0*SCREEN_W-20)/10/map->w;
	      double sy=(1.0*SCREEN_H-20)/10/map->h;
	      
	      PLAYER_SCALE=(sx<sy ? sx : sy);
	      if ((PLAYER_SCALE>0.95) && (PLAYER_SCALE<1.1)) PLAYER_SCALE=1;
	      
              row++; break; 
    }	// switch
    
    if (row==5) break;
  }
  
  fclose(f);
}


int StartGame(int,int);
int StartProboj(char *config)
{
  ConfOpen(config);
  nPlayers = ConfGetInt("nplayers");

  int nGames = ConfGetInt("ngames");
 
  for (int i=1; i<=nGames; i++) {
    if (StartGame(i,nGames) == RET_HALT) return RET_HALT;
    DELAY(10,0);
  } 
  
  return 1;
}

int StartGame(int map_number, int nGames)
{
  char map_name[1024], color_name[1024];
  sprintf(map_name,"game%d_map", map_number);

  fprintf(logfile, "\nNew game: %s\n",ConfGetStr(map_name));

  for (int i=0;i<NR_VAULTS;i++) vaults[i]=new KrResourceVault;
  
  LoadMap(ConfGetStr(map_name), &map, obj_sprites);

  // set default colors
  for (int pl=0; pl<MAX_PLAYERS; pl++) {
    player_color[pl][0] = default_player_color[pl][0];
    player_color[pl][1] = default_player_color[pl][1];
    player_color[pl][2] = default_player_color[pl][2];
  }
  
  sprintf(color_name,"game%d_color", map_number);
  char *colorify=ConfGetStr(color_name);

  if ((colorify!=NULL) && (strlen(colorify)>0)) {
    int col[3];
    sscanf(colorify,"%d,%d,%d",&(col[0]), &(col[1]), &(col[2]));
    for (int pl=0; pl<MAX_PLAYERS; pl++) {
      player_color[pl][0] = col[0];
      player_color[pl][1] = col[1];
      player_color[pl][2] = col[2];
    }
  }

  get_sprites(&engine, obj_sprites, text, vaults, ConfGetStr(map_name), &map);

  // start game: create BASE entities & sprites
  for (int pl=0; pl<nPlayers; pl++) {
    char sname[100]; sprintf(sname,"player%d",pl+1);
    char *exe = ConfGetStr(sname);
    
    Player[pl].exename=(char *)malloc(strlen(exe)+1); strcpy(Player[pl].exename,exe);
    Player[pl].seq=0;
    Player[pl].entCount=0;
    Player[pl].base=-1;
    
    Player[pl].sugar_sum = 0;
    Player[pl].kills = 0;

    // born BASE entity for this player
    born_entity(pl, etBase, map.bases[pl].x, map.bases[pl].y, obj_sprites);

  }
  DEBUG("Base sprites created\n");
  
  // create sugar sprites
  bzero(sugar_to_sprite,sizeof(sugar_to_sprite));

  for (int y=0; y<map.h; y++) 
   for (int x=0; x<map.w; x++)
    if (ID_SUGAR(map.field[x][y])) {
      sugar_to_sprite[(y<<8) | x] = create_sprite(x,y,1,5);		// x, y, frame number, z-depth
    }

  DEBUG("Sugar sprites created\n");

  // game main loop
  int round=0;
  int dx,dy;
  bool end;
  
  // clear text screen
  printf(CUR_MOVE_HOME);
  for (int i=0;i<100;i++) printf("                                                                              ");
  
  int n_step=0;
  
  while ((gui_getkey(&screen, dx, dy, end)) && (n_step < map.maxSteps)) {
    // if there are no more entities (except bases) and no sugar to buy new ones, stop the game
    int alive=0;
    for (int pl=0; pl<nPlayers; pl++) {
      if (Player[pl].entCount==1) {
        if (Player[pl].base<0) { alive++; continue; }		// no base, but 1 entity -> he lives! :)
	
	if ((Player[pl].ent[Player[pl].base].sugar >= map.prizeWorker) || (Player[pl].ent[Player[pl].base].sugar >= map.prizeFighter)) alive++;
      } else alive++;
    }
    if (alive==0) break;
    

    // print stats
    round++;
    
    // send DoStep event to all active entities (ie. having pending==0)
    fd_set fds;
    FD_ZERO(&fds);
    int max_fd=0;
    
//    DEBUG("Sending DoStep packets...\n");
    for (int pl=0; pl<nPlayers; pl++) {
      for (int i=0; i<Player[pl].entCount; i++) {
        TEntity *ent = &(Player[pl].ent[i]);

	// mark fd of the entity (also when pending==1 -- maybe the reply just wasn't fast enough..)
	FD_SET(ent->fd,&fds);
	if (ent->fd > max_fd) max_fd=ent->fd;	

	// if pending==0, send new DoSTep message
	if (ent->pending==0) {
//  	  printf("  -> to fd=%d (pl=%d, ent=%d)\n",ent->fd, pl+1, i);
          SendPacket(ent->fd, mkDoStepPacket(ent, &map));
	  ent->pending++;
	  ent->timeout=0;
	  ent->in_msgs=0;	// clear incoming messages
	} else {
	  ent->timeout++;
	  if (ent->timeout > MAX_TIMEOUT) {
//	    DEBUG("Timeout: 0x%08X, killing...\n",ent->id);
	    ent->zombie=1;
	  }
	}
      }
    }
  
  
    int sum_ent=0;
    char sysload[100];
    for (int pl=0; pl<nPlayers; pl++) sum_ent+=Player[pl].entCount;
    FILE *f=fopen("/proc/loadavg","r"); fscanf(f,"%s",sysload); fclose(f);

  
    // handle score updates
    if (n_step%10 == 0) ScoreCheckNewTerminal();
     
    // clear text screen
//    printf(CUR_MOVE_HOME);
//    for (int i=0;i<10;i++) printf("                                                                              ");
    
    printf(CUR_MOVE_HOME);
    printf("-------------------------------------------------------------\n");
    printf("Round %d, total %d agents, system load %s               \n",round,sum_ent,sysload);
    if (!in_intro) {
      ScoreSend("Game %d %d\n",map_number,nGames);
      ScoreSend("Players %d\n",nPlayers);
      ScoreSend("Round %d\n",round);
      ScoreSend("Agents %d\n",sum_ent);
    }
    
    for (int pl=0; pl<nPlayers; pl++) {
      int cnt1=0, cnt2=0, cnt3=0; int avg1=0;
      for (int i=0; i<Player[pl].entCount; i++) {
        switch ((Player[pl].ent[i].id & MASK_TYPE) >> SHIFT_TYPE) {
	  case etWorker: cnt1++; break;
	  case etFighter: cnt2++; break;
	}
	avg1+=Player[pl].ent[i].timeout;
      }
      
      
      cnt3=Player[pl].sugar_sum + 3*(cnt1*map.prizeWorker + cnt2*map.prizeFighter) + 
                                    Player[pl].kills/2;
      if (Player[pl].base>=0) cnt3 += 3*Player[pl].ent[Player[pl].base].sugar;

      Player[pl].score_game = cnt3;


      printf(" Player %d: %3d workers, %3d fighters, %5d sugar, score %d, lag %0.2lf     \n",pl+1,cnt1,cnt2,(Player[pl].base>=0 ? Player[pl].ent[Player[pl].base].sugar : -1 ),cnt3, 1.0*avg1/(cnt1+cnt2+1));


      if (!in_intro) {
        ScoreSend("PlayerName %d %s\n", pl+1, Player[pl].name); 
        ScoreSend("PlayerColor %d %d %d %d\n", pl+1, player_color[pl][0], player_color[pl][1], player_color[pl][2]);
        ScoreSend("PlayerWorkers %d %d\n", pl+1, cnt1);
        ScoreSend("PlayerFighters %d %d\n", pl+1, cnt2);
        ScoreSend("PlayerSugar %d %d\n", pl+1, (Player[pl].base>=0 ? Player[pl].ent[Player[pl].base].sugar : -1 ));
        ScoreSend("PlayerScoreGame %d %d\n", pl+1, Player[pl].score_game);
        ScoreSend("PlayerScore %d %d\n", pl+1, GlobalPlayer[pl].score + Player[pl].score_game);
        ScoreSend("PlayerAvgTimeout %d %g\n", pl+1, 1.0*avg1/(cnt1+cnt2+1));
        ScoreSend("PlayerKills %d %d\n", pl+1, Player[pl].kills );
      }
    }  
    printf("-------------------------------------------------------------\n");
    for (int i=0;i<10;i++) printf("                                                                              \n");




    // move all entity sprites
    for (int pl=0; pl<nPlayers; pl++) 
     for (int i=0; i<Player[pl].entCount; i++) {
       TEntity *ent = &(Player[pl].ent[i]);
       obj_sprites[ent->sprite]->SetPos(10+(int)(10*PLAYER_SCALE*ent->x+0.5), 10+(int)(10*PLAYER_SCALE*(map.h-1-ent->y)+0.5));
       if (ID_TYPE(ent->id)==etWorker) {
         if (ent->sugar) obj_sprites[ent->sprite]->SetFrame(5); else obj_sprites[ent->sprite]->SetFrame(4);
       }
     }

    // update color of sugar sprites
    for (int y=map.h-1; y>=0; y--)
      for (int x=0; x<map.w; x++) {
	int sugar = (map.field[x][y] & MASK_SUGAR) >> SHIFT_SUGAR;
	int sugar_sprite = sugar_to_sprite[(y<<8)|x];

	if (sugar_sprite>0) {
	  int gr = (int)(sugar*1.0/map.maxSugar*120) + 50;
	  
	  KrColorTransform color;
	  COLOR(gr,80, gr,80, gr,80);
	  obj_sprites[sugar_sprite]->SetColor(color);
	}
      }
    
    engine->Draw();


    // wait for answers from all & update map
    switch (fastmode) {
      case 1:  DELAY(0,100000);  break;
      case 2:  DELAY(0,50000);   break;
      case 3:  DELAY(0,25000);   break;
      case 4:  DELAY(0,10000);   break;
      case 0:  DELAY(0,1000);    break;
      case RET_HALT: return RET_HALT;   break;
    }
        
    struct timeval tv;
    tv.tv_sec = tv.tv_usec = 0;
    select(max_fd+1, &fds,NULL,NULL,&tv);	// now check from who we have an reply available

    // execute all actions we received, in random order
    TEntity *ent_list[MAX_PLAYERS * MAX_ENT];
    int n_ent=0;
    
    for (int pl=0; pl<nPlayers; pl++)
     for (int i=0; i<Player[pl].entCount; i++) {
       TEntity *ent = &(Player[pl].ent[i]);
       if (FD_ISSET(ent->fd, &fds)) ent_list[n_ent++]=ent;
     }
     
    // randomize the list and proces it
    for (int i=n_ent; i>0; i--) {
      int rn=random(i);
      TEntity *ent = ent_list[rn];
      ent_list[rn] = ent_list[--n_ent];

      if (ent->pending>0) ent->pending--;
      TPacket pac = RecvPacket(-ent->fd, packetEmpty);
	
      // process packet
      process_DoneStep(ent, ((ent->id & MASK_PLAYER) >> SHIFT_PLAYER)-1, pac, obj_sprites);
    } 

     
    // remove all zombies (i.e. not responding, killed, or crashed) entities
    for (int pl=0; pl<nPlayers; pl++)
     for (int i=0; i<Player[pl].entCount; ) {
       if (Player[pl].ent[i].zombie) {
         TEntity *ent = &(Player[pl].ent[i]);
	 
	 if (ID_TYPE(ent->id) == etBase) {
	   DEBUG("Base of player %d was destroyed!\n", ID_PLAYER(ent->id));
	   Player[pl].base = -1;
	 }
	 
         map.field[ent->x][ent->y] &= MASK_SUGAR;
	 
	 close(ent->fd);
	 kill(ent->pid,SIGKILL);
	 
	 delete_sprite(ent->sprite);
	 
	 Player[pl].ent[i] = Player[pl].ent[Player[pl].entCount-1];
	 Player[pl].entCount--;
	 
       } else i++;
     }
     waitpid(-1,NULL,WNOHANG);
     
     
     n_step++;
  }

  if (n_step>=map.maxSteps) DELAY(5,0);

  // kill all child processes
  for (int pl=0; pl<nPlayers; pl++)
   for (int i=0; i<Player[pl].entCount; i++) {
     close(Player[pl].ent[i].fd);
     kill(Player[pl].ent[i].pid,SIGKILL);
   }
  waitpid(-1,NULL,WNOHANG); 

  // remova _all_ sprites (incl. background?)
  for (int pl=0; pl<nPlayers; pl++)
   for (int i=0; i<Player[pl].entCount; i++) delete_sprite(Player[pl].ent[i].sprite);

  // remove all sugar sprites
    for (int y=map.h-1; y>=0; y--)
      for (int x=0; x<map.w; x++) {
	int sugar_sprite = sugar_to_sprite[(y<<8)|x];
	if (sugar_sprite>0) {
   	  delete_sprite(sugar_sprite);
 	  sugar_to_sprite[(y << 8) | x] = 0;
	}
      }
  

  delete_all_sprites();

  for (int pl=0; pl<nPlayers; pl++) {
    GlobalPlayer[pl].score += Player[pl].score_game;
  }

  // log
  fprintf(logfile,"  Game results:\n");
  for (int pl=0; pl<nPlayers; pl++) {
    fprintf(logfile, "    Player (%d) %-20s              %8d points\n",pl+1,Player[pl].name,Player[pl].score_game);
  }

  
  return 1;
}

// ---------------------------------------------------------------------------
int main(int /* argc */, char ** /* argv */)
{
  srand(time(NULL));

  obj_sprites = new KrSprite*[MAX_SPRITES];
  text = new KrTextBox*[MAX_PLAYERS+5];

  char s[1024];
  sprintf(s,"proboj.%d.log",(int)time(NULL));
  logfile = fopen(s,"w");
  fprintf(logfile, "Proboj Pocuvadlo 2004: CUKOR II\n");
  fprintf(logfile, "-------------------------------\n");


  DEBUG("calling init_SDL...\n");
  init_SDL(&screen, &engine);
  DEBUG("calling get_sprites\n");

  for (int i=0; i<MAX_PLAYERS; i++) GlobalPlayer[i].score=0;
  StartScoreTerminals(1058);
 

  fastmode=0; in_intro=1;
  StartProboj("proboj.intro.conf");


  for (int i=0; i<MAX_PLAYERS; i++) GlobalPlayer[i].score=0;
  fastmode=2; in_intro=0;
  StartProboj("proboj.conf");


  SDL_Quit();

  fprintf(logfile,"\n\n=========================================================================\nResults:\n\n");
  printf("\n\n=========================================================================\nResults:\n\n");
  for (int pl=0; pl<nPlayers; pl++) {
    printf( "Player (%d) %-20s              %8d points\n",pl+1,Player[pl].name,GlobalPlayer[pl].score);
    fprintf(logfile, "Player (%d) %-20s              %8d points\n",pl+1,Player[pl].name,GlobalPlayer[pl].score);
  }

  fclose(logfile);

  return 0;
}
