#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "config.h"
#include "comm.h"


char *PlayerName;

// born
//   called when the entity borns
//    ent          -- this entity (id, coordinates, ...)
//    nplayers     -- number of players in a game
//    map_w,map_h  -- coordinates, size of the map
extern int born(TEntity ent, int nplayers, int map_w, int map_h, int pr_worker, int pr_fighter);

// step_(base|fighter|worker)
//   make a single step of a given entity
//    ent          -- this entity (id, coordinates, ...)
//    map          -- part of map visible around entity (with other entities)
//
//    *out_msg     -- message to send
//    returns direction to move in
extern int step_base(TEntity ent, TMap map, TEntityCount cnt, TMessage *out_msg);
extern int step_fighter(TEntity ent, TMap map, TMessage *out_msg);
extern int step_worker(TEntity ent, TMap map, TMessage *out_msg);



int main(void)
{
  PlayerName = (char *)malloc(1024);
  strcpy(PlayerName, "(Noname)");

  // wait for born info
  TPacket pac = RecvPacket(FD,packetBorn);
  TEntity ent;    
  TMap map;
  TEntityCount cnt;
  int map_w, map_h, map_prWorker, map_prFighter;
    
  
  // parse born packet
  ent.id = GetInt(pac,0);
  map_w = GetByte(pac,4);
  map_h = GetByte(pac,5);
  ent.x = GetByte(pac,6);
  ent.y = GetByte(pac,7);
  int nplayers = GetByte(pac,8);
  ent.sugar = GetInt(pac,9);
  map_prWorker = GetInt(pac,13);
  map_prFighter = GetInt(pac,17);
  
  // call player's _born_ function
  born(ent, nplayers, map_w, map_h, map_prWorker, map_prFighter);

  // send name to master, if I am a base
  if (ID_TYPE(ent.id)==etBase) {
    SetInt (pac,0, ent.id);
    SetInt (pac,4, strlen(PlayerName)+1 );
    for (int i=0; i<=strlen(PlayerName); i++) SetByte(pac,8+i, PlayerName[i]);
    pac.packetType = packetName;
    
    // send DoneStep packet
    SendPacket(FD,pac);
  }
  
  
  // enter main event loop
  while (1) {
    // wait for DoStep packet
//    printf("%d: waiting for DoStep packet\n", getpid());
    pac = RecvPacket(FD,packetDoStep);
    
    // parse DoStep packet
    ent.id = GetInt(pac,0);
    ent.x = GetByte(pac,4);
    ent.y = GetByte(pac,5);
    ent.sugar = GetInt(pac,6);
    ent.in_msgs = GetByte(pac,10);
    int offset=11;

    // entity counts (for etBase, 0 otherwise)
    cnt.workers = GetByte(pac,offset++);
    cnt.fighters = GetByte(pac,offset++);
    
    // map
    for (int j=0; j<MAP_SIZE; j++)
     for (int i=0; i<MAP_SIZE; i++) {
       map.field[i][j] = GetInt(pac,offset);
       offset+=4;
     }
    
    // in messages
    for (int i=0; i<ent.in_msgs; i++) {
      ent.in_msg[i].id  = GetInt(pac,offset);
      ent.in_msg[i].msg = GetInt(pac,offset+4);
      offset+=8;
    }
    
    // call player's _step_xyz_ function
    pac.packetType=packetEmpty;
    TMessage out_msg; 
    int dir=dirStay;
    
    out_msg.id=0; out_msg.msg=0;
    
    switch ((ent.id & MASK_TYPE) >> SHIFT_TYPE) {
      case etBase:  
            dir = step_base(ent, map, cnt, &out_msg);
            break;

      case etFighter:  
            dir = step_fighter(ent, map, &out_msg);
            break;
		    
      case etWorker:  
    	    dir = step_worker(ent, map, &out_msg);
            break;
	    
      default:
            break;
    }
    
    // prepare DoneStep packet
    SetInt (pac,0, ent.id);
    SetByte(pac,4, dir & 0xFF);
    SetInt (pac,5, out_msg.id);
    SetInt (pac,9, out_msg.msg);
    pac.packetType = packetDoneStep;
    
    // send DoneStep packet
    SendPacket(FD,pac);
  }
  
  // finish
  return 0;
}
