#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdarg.h>
#include <signal.h>

#include "score.h"

static struct {
  int fd;
} ScoreTerminal[MAX_TERMINALS];

static int ScoreSocket=-1;
static int ScoreTerminalCount = 0;


int StartScoreTerminals(int port)
{
  int server_len;
  struct sockaddr_in server_address;
  
  ScoreSocket = socket(PF_INET, SOCK_STREAM, 6);
  if (ScoreSocket<0)
  {
    fprintf(stderr,"Error: unable to ceate socket! (errno=%d)\n",errno);
    return -1;
  }
  
  int v=1; 
  setsockopt(ScoreSocket,SOL_SOCKET,SO_REUSEADDR,&v,sizeof(v));
  fcntl(ScoreSocket,F_SETFD,FD_CLOEXEC);

  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = htonl(INADDR_ANY);
  server_address.sin_port = htons(port);
  server_len = sizeof(server_address);
  int res=bind(ScoreSocket, (struct sockaddr *)&server_address, server_len);
  if (res<0)
  {
    fprintf(stderr,"Error: unable to bind! (errno=%d)\n",errno);
    close(ScoreSocket);
    return -1;
  }

  res=listen(ScoreSocket, MAX_TERMINALS);
  if (res<0)
  {
    fprintf(stderr,"Error: unable to listen! (errno=%d)\n",errno);
    close(ScoreSocket);
    return -1;
  }

  printf("Socket initialized\n");
  return ScoreSocket;
}

int ScoreCheckNewTerminal(void)
{
  if ((ScoreSocket>=0) && (ScoreTerminalCount<MAX_TERMINALS)) {
    struct timeval tv;
    tv.tv_sec = tv.tv_usec = 0;
    
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(ScoreSocket,&fds);
    
    int res = select(ScoreSocket+1, &fds, NULL, NULL, &tv);
    if ((res==1) && (FD_ISSET(ScoreSocket,&fds))) {
      ScoreTerminal[ScoreTerminalCount].fd = accept(ScoreSocket,NULL,NULL);
      if (ScoreTerminal[ScoreTerminalCount].fd>=0) ScoreTerminalCount++;
    }
  }

  signal(SIGPIPE, SIG_IGN);
  
  return 1;
}

int ScoreSend(char *msg, ...)
{
  char s[1024];

  va_list ap;
  va_start(ap, msg);
  vsprintf(s, msg, ap);
  va_end(ap);
  
  
  for (int i=0; i<ScoreTerminalCount; i++) {
    if (ScoreTerminal[i].fd >= 0) {
      int len = strlen(s);
      int res = write(ScoreTerminal[i].fd, s, len);
      
      if (res<0) {
        ScoreTerminal[i] = ScoreTerminal[ScoreTerminalCount-1];
	ScoreTerminalCount--;
      }
    }
  }
  
  return 1;
}

