/*--License:
	Kyra Sprite Engine
	Copyright Lee Thomason (Grinning Lizard Software) 2001-2002
	www.grinninglizard.com/kyra
	www.sourceforge.net/projects/kyra

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	The full text of the license can be found in license.txt
*/


#include <SDL/SDL.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <Kyra/kyra.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>

#include "sprites/nina.h"
#include "sprites/charts.h"

#define SDL_TIMER_EVENT ( SDL_USEREVENT + 0 )
const int TIMER_INTERVAL = 40;

int SCREENX = 1024;
int SCREENY = 768;

//int SCREENX = 800;
//int SCREENY = 600;


#define MAX_PLAYERS		10
#define BUF_SIZE		1024*32

char buf[BUF_SIZE];

struct {
  char *name;
  int color[3];
  
  KrTextBox *text_name;
  KrTextBox *info[100];

  KrSprite *chart[2];

  int score_total, score_game;
} Player[MAX_PLAYERS];

int nPlayers = 0;
int Socket = -1;

int iGame=0, iRound=0;


KrTextBox *texts[100];
KrEngine *engine;
KrResourceVault *vault[10];



Uint32 TimerCallback(Uint32 interval)
{
	SDL_Event event;
	event.type = SDL_TIMER_EVENT;

	SDL_PeepEvents( &event, 1, SDL_ADDEVENT, 0 );
	return TIMER_INTERVAL;
}


int do_connect(char *addr, int port) {
  struct sockaddr_in dest_addr;

  // pripoj sa na server
  int fd = socket(AF_INET, SOCK_STREAM, 6); 

  dest_addr.sin_family = AF_INET;        
  dest_addr.sin_port = htons(port); 
  dest_addr.sin_addr.s_addr = inet_addr(addr);
  bzero(&(dest_addr.sin_zero), 8);    

  int res = connect(fd, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr));
  
  if (res == -1) {
    close(fd);
    printf("Unable to connect (errno=%d)!\n",errno);
    return -1;
  }

  fcntl(fd, F_SETFL, O_NONBLOCK);
  
  printf("Connected, fd=%d\n",fd);
  return fd;
}


KrTextBox * TextBox(KrFontResource *font, KrImNode *root, int L, int T, int W, int H,  double SX, double SY,  int R, int RR, int G, int GG, int B, int BB)
{
  KrTextBox *box = new KrTextBox( font, W, H, 0, KrTextBox::LEFT);

  box->SetPos(L,T);
  box->SetScale(SX,SY,KR_ALL_WINDOWS);

  KrColorTransform color;
  color.SetRed( R,RR );
  color.SetGreen( G,GG );
  color.SetBlue( B,BB );
  box->SetColor( color );

  engine->Tree()->AddNode(root, box);
  
  return box;
}


void UpdateCharts()
{
  int max[2] = {10000,10000};
  
  for (int pl=0; pl<nPlayers; pl++) {
    if (Player[pl].score_game > max[0]) max[0] = Player[pl].score_game;
    if (Player[pl].score_total > max[1]) max[1] = Player[pl].score_total;
  }
  
  
  int H=SCREENY/2-100;
  
  for (int pl=0; pl<nPlayers; pl++) {
    Player[pl].chart[0]->SetScale(3, (1.0*Player[pl].score_game/max[0])*H/19, KR_ALL_WINDOWS);
    Player[pl].chart[1]->SetScale(3, (1.0*Player[pl].score_total/max[1])*H/19, KR_ALL_WINDOWS);
  }
}


int main( int argc, char *argv[] )
{
  SDL_Surface* screen;
  char IP_ADDR[100];

  strcpy(IP_ADDR, "10.0.0.10");

  if (argc>1) {
    strcpy(IP_ADDR,argv[1]);
  }

  const SDL_version* sdlVersion = SDL_Linked_Version();
  if ( sdlVersion->minor < 2 )
  {
    printf( "SDL version must be at least 1.2.0" );
    exit( 254 );
  }

  /* Initialize the SDL library */
  if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE) < 0 ) {
    printf( "Couldn't initialize SDL: %s\n",SDL_GetError());
    exit(255);
  }

  /* Create a display for the image */
  screen = SDL_SetVideoMode( SCREENX, SCREENY, 32, SDL_HWSURFACE | SDL_ANYFORMAT | SDL_HWACCEL );
  if ( screen == NULL ) {
    exit(3);
  }

  engine = new KrEngine( screen );
  vault[0] = new KrResourceVault;

  SDL_WM_ToggleFullScreen(screen);

  engine->Draw();

  // code_b Load the dat file.
  // The dat file was carefully created in the sprite
  // editor. Loading allows us access to the 
  // MAGE, PARTICLE, and CARPET.
//	if ( !engine->Vault()->LoadDatFile( "tutorial1.dat" )) { printf( "Error loading the tutorial dat file\n" ); exit(255); }
  if ( !engine->Vault()->LoadDatFile( "sprites/nina.dat" )) { printf( "Error loading the nina dat file\n" ); exit(255); }
  if ( !vault[0]->LoadDatFile( "sprites/charts.dat" )) { printf( "Error loading the charts dat file\n" ); exit(255); }
//	if ( !engine->Vault()->LoadDatFile( "formula1_out.dat" )) { printf( "Error loading the nina dat file\n" ); exit(255); }

  // Create NINA font resource
  KrFontResource* ninaRes = engine->Vault()->GetFontResource( FONT_NINA );
  GLASSERT( ninaRes );

  // create textbox	
//  KrTextBox* text = new KrTextBox( ninaRes, 500,200,0, KrTextBox::LEFT);
//  text->SetPos(60,40);
//  text->SetScale(1.7,1.4,KR_ALL_WINDOWS);
//  text->SetTextChar("Ahoj oblak! :))", 0);
//  text->SetTextChar("VATAMA ABCDEFGHIJ", 1);

//  KrColorTransform colorXForm;
//  GlRandom random;
//  random.SetSeed(time(NULL));
//  colorXForm.SetRed( random.Rand( 128 )+100, 0 );
//  colorXForm.SetGreen( random.Rand( 128 )+50, 00 );
//  colorXForm.SetBlue( random.Rand( 128 )+50, 0 );
//  text->SetColor( colorXForm );
//
//  engine->Tree()->AddNode(0, text);

  // auticko
//  KrSpriteResource *carRes = engine->Vault()->GetSpriteResource(CARS_FORMULA); GLASSERT(carRes);
//  KrSprite *car = new KrSprite(carRes);
//  car->SetPos(SCREENX/3, SCREENY/2);
//  car->SetColor(colorXForm);
//	car->SetScale(0.7, 0.7);
//  engine->Tree()->AddNode(0, car);

  KrSpriteResource *chartRes = vault[0]->GetSpriteResource(CHARTS_BLANK); GLASSERT(chartRes);

	
  // background
  KrRGBA yellow;
  yellow.Set(0,0,20);
  engine->FillBackground(&yellow);

  SDL_Event event;
  bool done = false;

  // create common text boxes
  texts[0] = TextBox(ninaRes, NULL, 10,10,SCREENX-20,50, 0.9,0.9, 255,0,255,0,255,0 );
  texts[1] = TextBox(ninaRes, NULL, SCREENX-350,10,SCREENX-20,50, 0.9,0.9, 255,0,255,0,255,0 );

  texts[2] = TextBox(ninaRes, NULL, 10,SCREENY-35,SCREENX/2,50, 0.8,0.8, 255,0,255,0,255,0 );
  texts[3] = TextBox(ninaRes, NULL, SCREENX/2,SCREENY-35,SCREENX/2,50, 0.8,0.8, 255,0,255,0,255,0 );

  // Start timing!
  SDL_SetTimer( TIMER_INTERVAL, TimerCallback );

  for (int pl=0; pl<MAX_PLAYERS; pl++) {
    Player[pl].text_name = TextBox(ninaRes, NULL, -1000,-1000, SCREENX/3-10, SCREENY/2/2, 1,1, 255,0,255,0,255,0 ); 
    
    for (int i=0; i<5; i++) {
      Player[pl].info[2*i]   = TextBox(ninaRes, Player[pl].text_name, 10,40+30*i, 500,200, 0.7, 0.7, 255,0,255,0,255,0 );
      Player[pl].info[2*i+1] = TextBox(ninaRes, Player[pl].text_name, 180,40+30*i, 500,200, 0.7, 0.7, 255,0,255,0,255,0 );
      
      if (i==3) Player[pl].info[2*i]->SetScale(0.9, 0.9, KR_ALL_WINDOWS);
    }
    
    Player[pl].score_total = Player[pl].score_game = 0;
    
    for (int i=0; i<2; i++) {
      Player[pl].chart[i] = new KrSprite(chartRes);
      Player[pl].chart[i]->SetScale(3,2,KR_ALL_WINDOWS);
      Player[pl].chart[i]->SetPos(-1000, -1000);
      engine->Tree()->AddNode(NULL, Player[pl].chart[i]);
    }
  }

  texts[0]->SetTextChar("Proboj Pocuvadlo 2004: CUKOR II.", 0);
  texts[2]->SetTextChar("This game score", 0);
  texts[3]->SetTextChar("Total score", 0);
  char s[1024];


  while( !done && SDL_WaitEvent(&event) )
  {
    if ( event.type == SDL_QUIT ) break;

    // check incoming data on socket
    if (Socket<0) Socket = do_connect(IP_ADDR,1058);
    
    if (Socket>=0) {
      int len = read(Socket,buf,BUF_SIZE);
      char *q = buf;
      char *qq;
      
      if (len>0) {
//        printf("IN: (len=%d, errno=%d) %s\n",len,errno,buf);
        q[len]=0;
	
	while ((qq=strchr(q,'\n'))!=NULL) {
	  char cmd[1024];
	  int pos,cnt;
	  
	  *qq=0;
	  cnt = sscanf(q,"%s %n",cmd,&pos);
	  if (cnt==1) {
	    q+=pos;
	    if (!strcmp(cmd,"Game")) {
	      sscanf(q,"%d %*d",&iGame);
	      sprintf(s,"Game %d, round %d",iGame,iRound);
	      texts[1]->SetTextChar(s,0);
	    }
	    
	    if (!strcmp(cmd,"Players")) {
	      sscanf(q,"%d",&nPlayers);
	      for (int i=0; i<nPlayers; i++) {
	        Player[i].text_name->SetPos(10+(i/2)*SCREENX/3, 50+((i%2)*SCREENY/2/2));
	        Player[i].chart[0]->SetPos(SCREENX/13*i + 10, SCREENY-40);
	        Player[i].chart[1]->SetPos(SCREENX/13*i + SCREENX/2, SCREENY-40);
	      }
	      
	      for (int i=nPlayers; i<MAX_PLAYERS; i++) {
	        Player[i].text_name->SetPos(-1000, -1000);
	        Player[i].chart[0]->SetPos(-1000, -1000);
	        Player[i].chart[1]->SetPos(-1000, -1000);
	      }
	    }
	    
	    if (!strcmp(cmd,"Round")) {
	      sscanf(q,"%d %*d",&iRound);
	      sprintf(s,"Game %d, round %d",iGame,iRound);
	      texts[1]->SetTextChar(s,0);
	    }
	    
	    if (!strcmp(cmd,"Agents")) ; // printf("Agents: %s\n",q);


	    if (!strcmp(cmd,"PlayerName")) {
	      int pl, len;
	      sscanf(q,"%d %n", &pl, &len);  pl--; q+=len;
	      sprintf(s,"%s",q);
	      Player[pl].text_name->SetTextChar(s, 0);
	    }


	    if (!strcmp(cmd,"PlayerColor")) {
	      int pl, r,g,b;
	      sscanf(q,"%d %d %d %d", &pl, &r, &g, &b);  pl--;

	      KrColorTransform color;
	      color.SetRed( r,0 );
	      color.SetGreen( g,0 );
	      color.SetBlue( b,0 );
	      Player[pl].text_name->SetColor(color);
	      Player[pl].chart[0]->SetColor(color);
	      Player[pl].chart[1]->SetColor(color);
	      
//	      color.SetRed(255,0); color.SetGreen(255,0); color.SetBlue(255,0);
//	      for (int i=0; i<10; i++) Player[pl].info[i]->SetColor(color);
	    }

	    if (!strcmp(cmd,"PlayerSugar")) {
	      int pl, n;
	      sscanf(q,"%d %d ", &pl, &n);  pl--;
	      sprintf(s,"Sugar: %d",n);
	      Player[pl].info[0]->SetTextChar(s, 0);
	    }
	    
	    if (!strcmp(cmd,"PlayerAvgTimeout")) {
	      int pl;  double timeout;
	      sscanf(q,"%d %lf", &pl, &timeout);  pl--;
	      sprintf(s,"Lag: %0.2lf",timeout);

	      Player[pl].info[1]->SetTextChar(s,0);
	    }

	    if (!strcmp(cmd,"PlayerWorkers")) {
	      int pl, n;
	      sscanf(q,"%d %d ", &pl, &n);  pl--;
	      sprintf(s,"Workers: %d",n);
	      Player[pl].info[2]->SetTextChar(s, 0);
	    }

	    if (!strcmp(cmd,"PlayerFighters")) {
	      int pl, n;
	      sscanf(q,"%d %d ", &pl, &n);  pl--;
	      sprintf(s,"Fighters: %d",n);
	      Player[pl].info[3]->SetTextChar(s, 0);
	    }

	    if (!strcmp(cmd,"PlayerScore")) {
	      int pl, score;
	      sscanf(q,"%d %d ", &pl, &score);  pl--;
	      sprintf(s,"Total: %d",score);
	      Player[pl].info[4]->SetTextChar(s, 0);
	      Player[pl].score_total = score;
	      UpdateCharts();
	    }

	    if (!strcmp(cmd,"PlayerKills")) {
	      int pl, score;
	      sscanf(q,"%d %d ", &pl, &score);  pl--;
	      sprintf(s,"Kills: %d",score);
	      Player[pl].info[5]->SetTextChar(s, 0);
	    }

	    if (!strcmp(cmd,"PlayerScoreGame")) {
	      int pl, score;
	      sscanf(q,"%d %d ", &pl, &score);  pl--;
	      sprintf(s,"Score: %d",score);
	      Player[pl].info[6]->SetTextChar(s, 0);
	      Player[pl].score_game = score;
	      UpdateCharts();
	    }

	  }
	  q=qq+1;
	}
      } else if ((len==-1) && (errno!=EAGAIN)) {
        printf("Socket read error, disconnecting...\n");
	close(Socket);
	Socket=-1;
      }
    }


    switch(event.type) {
	case SDL_KEYDOWN: {
		switch (event.key.keysym.sym) {
			case SDLK_ESCAPE: 
					done=true; 
					break;
					
			case SDLK_SPACE:
//					colorXForm.SetRed( random.Rand( 128 )+100, 0 );
//					colorXForm.SetGreen( random.Rand( 128 )+50, 00 );
//					colorXForm.SetBlue( random.Rand( 128 )+50, 0 );
//					car->SetColor(colorXForm);
					break;
			case SDLK_f:
					SDL_WM_ToggleFullScreen(screen);
					break;
			case SDLK_q:
					done=true;
					break;
					
			default: 
					break;
		}
	    }
	    break;

	case SDL_TIMER_EVENT: {
		// code_b Move the carpet.
//		car->DoStep();
//		car->SetPos((car->X()<1?SCREENX:car->X()-1), car->Y());
		engine->Draw();
	    }
	    break;

	default:
	    break;
    }
    
  }

  delete engine;

  SDL_Quit();	
  return 0;
}

